/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors: Ivan Kurmanov, Jasper van de Gronde (win32 port)
 */

#include <string.h>
#include <stdio.h>
#include "../utils/utils.h"
#include <pthread.h>
#ifdef WIN32
#include <windows.h>
#undef INSERT
#undef SELECT
#undef DELETE
#undef CREATE
#undef UPLOAD
#undef UNLOAD
#undef ENTER
#undef LEAVE
#undef DROP
#endif
#include "queryParser.h"
#include <pthread.h>

QueryParserFunc parserList[10];

#ifdef WIN32
#else
#define isspace(x) ((isspace(x) || (x == '\x0D') || (x == '\x0A') || (x == '\r')))
#endif

#define DEBUG_QUERY_PARSER 0

#if DEBUG_QUERY_PARSER
#define DEBUG if(1)
#else
#define DEBUG if(0)
#endif

QueryParseResult ParseInsert(Query nq);
QueryParseResult ParseDelete(Query nq);
QueryParseResult ParseSelect(Query nq);
QueryParseResult ParseCreate(Query nq);

void* parseTermOfType (char* term, RDF_ValueType type) ;
RDF_ValueType termType (const char *term) ;


int gQueryCounter = 0;
pthread_mutex_t gQueryCounterMutex = PTHREAD_MUTEX_INITIALIZER;

Query gQueryPool[4096];

int
getNextQueryCounter () {
  int ans;
  pthread_mutex_lock(&gQueryCounterMutex);
  ans = gQueryCounter++;
  pthread_mutex_unlock(&gQueryCounterMutex);
  return ans;
}


void
freeQuery (Query nq) {
  Literal l = nq->literals;
  Variable v = nq->variables;
  while (v) {
    Variable next = v->next;
    freeMem((void*)v->name);
    freeMem((void*)v);
    v = next;
  }
  while (l) {
    Literal next = l->next;
    freeMem((void*)l->source);
    freeMem((void*)l->arc);
    freeMem((void*)l->target);
    l = next;
  }
  freeMem((void*)nq->database);
  freeMem((void*)nq->url);
  freeMem((void*)nq->queryString);
  destroy_constraint_list(nq->constraint_list);
  destroy_for_list(nq->for_list);
  freeMem((void*)nq);
}

QueryType
getQueryType (const char* query) {
  if (startsWith_forgiving_ws(query, "select")) {
    return SELECT;
  } else if (startsWith_forgiving_ws(query, "load")) {
    return UPLOAD;
  } else if (startsWith_forgiving_ws(query, "unload")) {
    return UNLOAD;
  } else if (startsWith_forgiving_ws(query, "enter")) {
    return ENTER;
  } else if (startsWith_forgiving_ws(query, "drop")) {
    return DROP;
  } else if (startsWith_forgiving_ws(query, "leave")) {
    return LEAVE;
  } else if (startsWith_forgiving_ws(query, "insert")) {
    return INSERT;
  } else if (startsWith_forgiving_ws(query, "delete")) {
    return DELETE;
  } else if (startsWith_forgiving_ws(query, "create")) {
    return CREATE;
  } else return UNKNOWN_QUERY_TYPE;
}


QueryParseResult
parseQuery (const char* queryString, Query *result) {

  QueryType type = getQueryType(queryString);
  Query nq ;
  QueryParseResult err;

  if (!type) {
    return UNKNOWN_QUERY_TYPE_ERROR;
  }

  nq =  (Query)getMem(sizeof(QueryStruct));
  bzero(nq, sizeof(QueryStruct));
  nq->type = type;
  nq->queryString = strdup(queryString);

  if (0 != (err = (*(parserList[type]))(nq))) {
    freeQuery(nq);
    return err;
  }

  nq->id = getNextQueryCounter();
  gQueryPool[nq->id] = nq;
  *result = nq;

  return 0;
}
  
  
/***   Next Word   ***/  
  
size_t 
nextWord (const char* str, size_t* start) {
  size_t end;
  int literalp;

  while ((str[*start] != 0) && (isspace(str[*start]) || (str[*start] == ','))) {
    (*start)++;
  }

  end = *start;

  if (str[end] == 0) return 0;

  literalp = (str[end] == '(');

  while ((str[end] != 0) && (literalp ? (str[end] != ')') : !isspace(str[end]))) {
    end++;
  }

  if (str[end] == ')') end++;
  return end;
}


/*****  IKu:  **********************/
/***     Next Word or string     ***/
size_t 
nextWord_or_string (const char* str, size_t* start) {
  size_t end;
  int inquote;
  char last;

  DEBUG { printf("in nextWord_or_string(), looking at '%s'\n", &str[*start] ); }

  while ((str[*start] != 0) && (isspace(str[*start]) || (str[*start] == ','))) {
    (*start)++;
  }

  end = *start;

  DEBUG { printf("in nextWord_or_string(), skipped ws '%s'\n", &str[*start] ); }

  if (str[end] == 0) return 0;

  inquote = (str[end] == '\'');

  while (1) {
    DEBUG 
      printf("in nextWord_or_string(), main cycle at '%s'\n", &str[end] );
    last = str[end];
    end++;
    if( str[end] == 0 ) { break; }
    if( inquote ) {
      if( str[end] == '\'' ) {
        if( last != '\\' ) { break; }
      }
    } else {
      if ( isspace( str[end] ) || (str[end] == ')') ) { break; }
    }
  }

  if (str[end] == '\'') end++;

  DEBUG 
    printf("in nextWord_or_string(), returning '%i'\n", end );

  return end;
}
/***************/


/*** Iku:  Next Triple   ***/

Literal 
next_triple( const char * qstring, size_t word_start, size_t* end ) {
  size_t word_end ;

  Literal ans;
  char *source, *arc,  *temp;
  void* target;
  int index = 0;
  //  int start;
  RDF_ValueType type;
  //  int sttrg = 0;
  source = arc = target = temp = 0;
  
  DEBUG 
    printf("in next_triple(), looking for '('\n" );

  //  find the first - openning parenthesis and skip it
  
  while (  isspace( qstring[word_start] ) || ( qstring[word_start] == ',' ) ) {
    word_start++;
    if ( qstring[word_start] == 0) {  return 0;  }
  }
          
  if( qstring[word_start] != '(' ) { 
    return (void*)1;
  } 
  word_end = word_start+1;

  DEBUG 
    printf("in next_triple(), looking for arc\n" );

  //  find the first part of a triple: property or arc

  word_start = word_end;
  word_end = nextWord_or_string( qstring, &word_start );
  if( !word_end ) { return (void*)1; }

  //  save it

  temp = getMem( word_end -word_start+1 );
  bzero( temp, word_end - word_start+1 );
  memcpy( temp, &qstring[word_start], word_end-word_start );
  arc = temp;

  DEBUG 
    printf("in next_triple(), looking for source\n" );

  //  find the second part of a triple: subject or source

  word_start = word_end;
  word_end = nextWord_or_string( qstring, &word_start );
  if( !word_end ) { return (void*)1; }

  //  save it

  temp = getMem( word_end -word_start+1 );
  bzero( temp, word_end - word_start+1 );
  memcpy( temp, &qstring[word_start], word_end-word_start );
  source = temp;
  
  DEBUG 
    printf("in next_triple(), looking for target\n" );

  //  find the third part of a triple: value

  word_start = word_end;
  word_end = nextWord_or_string( qstring, &word_start );
  if( !word_end ) { return (void*)1; }

  //  save it

  temp = getMem( word_end -word_start+1 );
  bzero( temp, word_end - word_start+1 );

  if( qstring[word_start] == '\'' ) { 
    int counter= 0;
    int slash = 0;
    int src_bytes = word_end - word_start;
    size_t destination=0;
    while( counter < src_bytes ) {
      char c = qstring[word_start+counter];
      counter++;

      if( slash ) { temp[destination++] = c; slash=0;  }
      else {
        if( c == '\\' ) { slash = 1; }
        else 
          if( c=='\t' ) { temp[destination++]=' '; } 
          else 
            {  temp[destination++] = c;   }
      }
    }
  } else {
    memcpy( temp, &qstring[word_start], word_end-word_start );
  }

  type = termType(temp);
  target = parseTermOfType(temp, type);
  freeMem(temp);
  
  DEBUG 
    printf("in next_triple(), looking for ')'\n" );

  //  find the last - closing parenthesis and skip it
  
  word_start = word_end;

  while (  isspace( qstring[word_start] ) || ( qstring[word_start] == ',' ) ) {
    word_start++;
    if ( qstring[word_start] == 0) {  return (void*)1;  }
  }
          
  if( qstring[word_start] != ')' ) { 
    return (void*)1;
  } 
  word_end = word_start+1;

  (*end) = word_end;
  
  DEBUG 
    printf("in next_triple(), resulting structure\n" );
  
  //  generate the resulting structure
  
  ans = (Literal)getMem(sizeof(LiteralStruct));
  bzero(ans, sizeof(LiteralStruct));
  ans->source = source;
  ans->arc    = arc;
  ans->type   = type;
  ans->target = target;


  return ans;
}

/*** end ***/



size_t 
nextVariable (const char* str, size_t* start) {
  size_t end;


  while ((str[*start] != 0) && (isspace(str[*start]) || (str[*start] == ','))) {
    (*start)++;
  }

  end = *start;
  if (str[end] == 0) return 0;
  while ((str[end] != 0) && (!isspace(str[end])) && (str[end] != ',')) {
    end++;
  }
  return end;
}
  
int
checkTerm (const char *term) {
  int len = strlen(term);
  if ((term[0] == '\'') && (term[len-1] != '\'')) return 1;
  if ((term[0] != '\'') && (term[len-1] == '\'')) return 1;
  return 0;
}


Literal
parseLiteral (const char* str, size_t len) {
#define PLE  {freeMem((void*)arc); freeMem((void*)source); return 0;}

  Literal ans;
  char *source, *arc,  *temp;
  void* target;
  size_t index = 0;
  size_t start;
  RDF_ValueType type;
  int sttrg = 0;
  source = arc = target = temp = 0;

  while ((index < len) && ((str[index] == '(') || (isspace(str[index])))) index++;

  
  start = index;
  while ((index < len) && (!isspace(str[index]))) index++;
  if ((index == start) || (index == len)) {
    return 0;
  }
  temp = getMem(index-start+1);
  bzero(temp, index-start+1);
  memcpy(temp, &str[start], index-start);
  arc = temp;

  while ((index < len) && (isspace(str[index]))) index++;
  start = index;
  while ((index < len) && (!isspace(str[index]))  && (str[index] != ')')) index++;
  if ((index == start) || (index == len)) {PLE}
  
  if (checkTerm(arc)) {PLE}

  temp = getMem(index-start+1);
  bzero(temp, index-start+1);
  memcpy(temp, &str[start], index-start);
  source = temp;
  if (checkTerm(source)) {PLE}

  while ((index < len) && (isspace(str[index]))) index++;
  start = index;
  sttrg = (str[start] == '\'');
  if (sttrg) index++;
  while ((index < len) && (sttrg ? (str[index] != '\'') : 
                           (!isspace(str[index])) && (str[index] != ')'))) index++;
  if (str[index] == '\'') index++;
  if ((index == start) || (index == len)) {PLE}
  
  temp = getMem(index-start+1);
  bzero(temp, index-start+1);
  memcpy(temp, &str[start], index-start);
  if (checkTerm(temp)) {PLE}
  type = termType(temp);
  target = parseTermOfType(temp, type);
  freeMem(temp);


  ans = (Literal)getMem(sizeof(LiteralStruct));
  bzero(ans, sizeof(LiteralStruct));
  ans->source = source;
  ans->arc    = arc;
  ans->type   = type;
  ans->target = target;

  return ans;
}


/***   Parse Insert   ***/

QueryParseResult 
ParseInsert (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "insert", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "into", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  while (1) {
    Literal l;
    
    DEBUG { printf("before next_triple(), qstring: %s\n", &nq->queryString[start] ); }
    l = next_triple( nq->queryString, start, &end );
    DEBUG { printf("after next_triple(), end-start: %i ; end: %i\n", end-start, end ); }

    if (l == (void*)1) return MALFORMED_LITERAL_ERROR;
    if (end-start > 0) {
      // parseLiteral(&nq->queryString[start], end-start);
      l->next = nq->literals;
      nq->literals = l;
      start = end;
    } else { 
      break;
    }
  }
  
  return NO_PARSE_ERROR;
}

int 
verifyVars (Query nq) {
  Variable v = nq->variables;
  while (v) {
    if (!v->verified && v->outputVariablep) {
      printf("Unconstrained variable : %s\n", v->name);
      return 1;
    } 
    v = v->next;
  }
  return 0;
}
  

/***   Add Literal Vars   ***/

void
addLiteralVars (Query nq, const char* str) {
  if (str[0] == '?') {
    Variable v = nq->variables;
    while (v) {
      if (streq(v->name, str)) {
        v->verified = 1;
        return;
      }
      v = v->next;
    }
    v = (Variable)getMem(sizeof(VariableStruct));
    v->name = strdup(str);
    v->next = nq->variables;
    nq->variables = v;
  }
}


void
addVariableToQuery (Query nq, Variable v) {
  Variable vn = nq->variables;
  Variable prev = vn;

  while (vn) {
    prev = vn;
    vn = vn->next;
  }
  
  if (!prev) {
    nq->variables = v;
  } else {
    prev->next = v;
  }
}

int parse_where_statement ( Query nq, size_t end, size_t start );

/***   Parse Select   ***/

QueryParseResult 
ParseSelect (Query nq) {

  size_t  start = 0;
  size_t  end = 0;
  int err;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "select", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  while (end) {
    Variable v;
    end = nextVariable(nq->queryString, &start);
    if (end == 0) return GENERAL_SYNTAX_ERROR;
    if (strncasecmp(&nq->queryString[start], "from", end-start) == 0) {
      start = end;
      break;
    }
    if (nq->queryString[start] != '?') {
      return GENERAL_SYNTAX_ERROR;
    }
    
    v = (Variable)getMem(sizeof(VariableStruct));
    bzero(v, sizeof(VariableStruct));
    v->name = (char*)getMem(end-start+1);
    bzero(v->name, end-start+1);    
    memcpy(v->name, &nq->queryString[start], end-start);
    addVariableToQuery(nq, v);
    v->outputVariablep = 1;
    start = end;
  }

 
  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "where", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  
  nq->outputType = TAB_LIMITED_OUTPUT;


  if( err= parse_where_statement( nq, end, start ) ) {
    return err;
  }

  return NO_PARSE_ERROR;
}



int parse_where_statement ( Query nq, size_t end, size_t start ) {

  while (end) {
    Literal l;
    
    DEBUG
      printf("before next_triple(), qstring: %s\n", &nq->queryString[start] );
    l = next_triple( nq->queryString, start, &end );
    DEBUG 
      printf("after next_triple(), end-start: %i ; end: %i, returned: %i\n", 
             end-start, 
             end,
             (long) l );
    
    if ( l == (void*)1 ) {
      
      end = nextWord(nq->queryString, &start);
      if (strncasecmp(&nq->queryString[start], "output", end-start) == 0) {
        DEBUG 
          printf ( "'output' found\n" );

        start = end;
        end = nextWord(nq->queryString, &start);
        if (strncasecmp(&nq->queryString[start], "rdf-xml", end-start) == 0) {
          nq->outputType = RDF_XML_OUTPUT;
        } else if (strncasecmp(&nq->queryString[start], "javascript", end-start) == 0) {
          nq->outputType = JAVASCRIPT_OUTPUT;
        } else if (strncasecmp(&nq->queryString[start], "tab-limited", end-start) == 0) {
          nq->outputType = TAB_LIMITED_OUTPUT;
        }  else if (strncasecmp(&nq->queryString[start], "variable-list", end-start) == 0) {
          nq->outputType = VAR_LIST_OUTPUT;
        }
        break;
        
      } else if (strncasecmp(&nq->queryString[start], "and", end-start) == 0) {
        /* Jasper van de Gronde
         * Parse the constraint list.
         */

        DEBUG 
          printf ( "'and' found\n" );

        start = end;

        {
          struct state_t state;
          state.querystart = nq->queryString;
          state.queryend = nq->queryString + strlen(nq->queryString);
          state.curstart = nq->queryString + start;
          state.curend = state.curstart;
          nq->constraint_list = parse_constraint_list(&state);
          start = state.curstart - nq->queryString;
        }

      } else if (strncasecmp(&nq->queryString[start], "using", end-start) == 0) {
        /* Jasper van de Gronde
         * Parse the namespace list.
         */

        DEBUG 
          printf ( "'using' found\n" );

        start = end;

        {
          struct state_t state;
          state.querystart = nq->queryString;
          state.queryend = nq->queryString + strlen(nq->queryString);
          state.curstart = nq->queryString + start;
          state.curend = state.curstart;
          nq->for_list = parse_for_list(&state);
          start = state.curstart - nq->queryString;
        }

      } else if ( end != 0 ) {
        return MALFORMED_LITERAL_ERROR;
      }
    } else if( l== (void*) 0 ) {
      break;
    } else if (end-start > 0) {
      // parseLiteral(&nq->queryString[start], end-start);

      addLiteralVars(nq, l->arc);
      addLiteralVars(nq, l->source);
      if (l->type == RDF_RESOURCE_TYPE) addLiteralVars(nq, l->target);

      l->next = nq->literals;
      nq->literals = l;
      start = end;

    } else { 
      break;
    }
  }
  
  DEBUG 
    printf( "at the end of parse_where_statement()\n" );

  if (!nq->literals) return MALFORMED_LITERAL_ERROR;
  if (verifyVars(nq)) return RDF_UNCONSTRAINED_VARIABLE;
  DEBUG 
    printf( "after verifyVars(nq) in parse_where_statement(), before exit\n" );

  return 0;

}




/***   Parse Delete   ***/

QueryParseResult 
ParseDelete(Query nq) {
  size_t  start = 0;
  size_t  end = 0;
  int delete_where_mode =0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "delete", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "from", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!strncasecmp(&nq->queryString[start], "where", end-start)) { 
    delete_where_mode = 1; 
    start = end;
  }

  if( delete_where_mode ) {
    int err;
    Variable var ;

    nq -> type = DELETEWHERE;
    nq -> outputType = 0;
    if( err= parse_where_statement( nq, end, start ) ) {
      return err;
    }

    var = nq->variables;
    while (var) {
      var -> outputVariablep = 1;
      var = var->next;
    }
    
  } else {

  while (end) {
    Literal l;
      
      DEBUG { printf("before next_triple(), qstring: %s\n", &nq->queryString[start] ); }
      l = next_triple( nq->queryString, start, &end );
      DEBUG { printf("after next_triple(), end-start: %i ; end: %i\n", end-start, end ); }
      
      if (l == (void*)1) return MALFORMED_LITERAL_ERROR;
      if (end-start > 0) {
        // parseLiteral(&nq->queryString[start], end-start);
      l->next = nq->literals;
      nq->literals = l;
      start = end;
        
      } else { 
        break;
    }
  }
  
  }
 
    /****
    ****/
  
  
  return NO_PARSE_ERROR;
}


QueryParseResult 
ParseCreate (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "create", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "database", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

QueryParseResult 
ParseDrop (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "drop", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "database", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

QueryParseResult 
ParseEnter (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "enter", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "namespace", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  if (strncasecmp(&nq->queryString[start], "xmlns", 5)) return GENERAL_SYNTAX_ERROR;
  nq->namespacePrefix = (char*)getMem(end-start+1);
  bzero(nq->namespacePrefix, end-start+1);
  memcpy(nq->namespacePrefix, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  nq->namespace = (char*)getMem(end-start+1);
  bzero(nq->namespace, end-start+1);
  memcpy(nq->namespace, &nq->queryString[start], end-start);
  start = end;
  
  return NO_PARSE_ERROR;
}

QueryParseResult 
ParseLeave (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "leave", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "namespace", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (!end) return GENERAL_SYNTAX_ERROR;
  if (strncasecmp(&nq->queryString[start], "xmlns", 5)) return GENERAL_SYNTAX_ERROR;
  nq->namespacePrefix = (char*)getMem(end-start+1);
  bzero(nq->namespacePrefix, end-start+1);
  memcpy(nq->namespacePrefix, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end)   return GENERAL_SYNTAX_ERROR;
  return NO_PARSE_ERROR;
  
}



QueryParseResult 
ParseUpload (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "load", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "rss", end-start) == 0) {
    nq->fileType = RSS;
  } else if (strncasecmp(&nq->queryString[start], "rdf_xml", end-start) == 0) {
    nq->fileType = RDF_XML;
  } else if (strncasecmp(&nq->queryString[start], "dmoz_rdf", end-start) == 0) {
    nq->fileType = DMOZ_RDF_XML;
  } else {
    return RDF_UNKNOWN_FILE_FORMAT_ERROR;
  }

  start = end;

  /* Jasper van de Gronde
     The if below is commented out because it should be able to load files.
     Also, if this word is read here the next word will be "into" instead of the url.
  */
  //end = nextWord(nq->queryString, &start);
  //if (strncasecmp(&nq->queryString[start], "file", end-start)) return GENERAL_SYNTAX_ERROR;
  //start = end;

  end = nextWord(nq->queryString, &start);
  nq->url = (char*) getMem(end-start+1);
  memcpy(nq->url, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "into", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end != 0) return GENERAL_SYNTAX_ERROR;

  return NO_PARSE_ERROR;
}


QueryParseResult 
ParseUnload (Query nq) {
  size_t  start = 0;
  size_t  end = 0;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "unload", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "file", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->url = (char*) getMem(end-start+1);
  memcpy(nq->url, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (strncasecmp(&nq->queryString[start], "from", end-start)) return GENERAL_SYNTAX_ERROR;
  start = end;

  end = nextWord(nq->queryString, &start);
  nq->database = (char*)getMem(end-start+1);
  bzero(nq->database, end-start+1);
  memcpy(nq->database, &nq->queryString[start], end-start);
  start = end;

  end = nextWord(nq->queryString, &start);
  if (end != 0) return GENERAL_SYNTAX_ERROR;

  return NO_PARSE_ERROR;

}


void
InitQueryParser () {
  parserList[INSERT] = ParseInsert;
  parserList[SELECT] = ParseSelect;
  parserList[DELETE] = ParseDelete;
  parserList[CREATE] = ParseCreate;
  parserList[UPLOAD] = ParseUpload;
  parserList[UNLOAD] = ParseUnload;
  parserList[ENTER]  = ParseEnter;
  parserList[LEAVE]  = ParseLeave;
  parserList[DROP]  =  ParseDrop;
}

void
describeQuery (Query q) {
  int n = 1;
  Literal lit = q->literals;
  Variable var = q->variables;
  printf("Optype = %i\nDB=%s\n", q->type, q->database);
  if (lit) {
    printf("Literals:\n");
    while (lit) {
      printf("%i) source=%s, arc=%s, target=%s\n", n++, lit->source, lit->arc, lit->target);
      lit = lit->next;
    }
  }
  if (var) {
    n=1;
    printf("Variables:\n");
    while (var) {
      printf("%i) name=%s\n", n++, var->name);
      var = var->next;
    }
  }
}

/*
void 
testParser () {
  Query ans;
  char line[1024];
  for (;;) {
    char *l;
    QueryParseResult err;
    printf("\n>");
    fflush(stdout);
    bzero(line,  1023);
    l = gets(line);
    if (!nullString(line, strlen(line))) {
      ans = 0;
      ProcessQuery(l, &ans);    
    }
  }
}
*/
